@@ -32,11 +32,19 @@ using SymbolicUtils
 using SymbolicUtils: BasicSymbolic, exprtype
 using SymbolicUtils: SYM, TERM, ADD, MUL, POW, DIV
 
+"""
+    is_exp(expr)
+
+Check if an expression is of the form `exp(...)`.
+
+Returns `true` if `expr` is a symbolic expression with the exponential function as its operation,
+`false` otherwise.
+"""
 is_exp(expr) = false
 is_exp(expr::BasicSymbolic) = exprtype(expr) == TERM && operation(expr) == exp
 
 # unused. This function provides a measure of the "size" of an expression, for use in proofs
-# of termination and debugging nontermintion only:
+# of termination and debugging nontermination only:
 # function S(expr, x)
 #     expr === x && return Set([x])
 #     expr isa BasicSymbolic || return Set([])
@@ -55,17 +63,42 @@ is_exp(expr::BasicSymbolic) = exprtype(expr) == TERM && operation(expr) == exp
 
 Compute the limit of `expr` as `x` approaches infinity and return `(limit, assumptions)`.
 
-This is the internal API boundry between the internal limits.jl file and the public
+This is the internal API boundary between the internal limits.jl file and the public
 SymbolicLimits.jl file
 """
-function limit_inf(expr, x::BasicSymbolic{Field}) where Field
+function limit_inf(expr, x::BasicSymbolic{Field}) where {Field}
     assumptions = Set{Any}()
     limit = signed_limit_inf(expr, x, assumptions)[1]
     limit, assumptions
 end
 
-signed_limit_inf(expr::Field, x::BasicSymbolic{Field}, assumptions) where Field = expr, sign(expr)
-function signed_limit_inf(expr::BasicSymbolic{Field}, x::BasicSymbolic{Field}, assumptions) where Field
+"""
+    signed_limit_inf(expr, x, assumptions)
+
+Compute the limit of `expr` as `x` approaches infinity and return `(limit, sign)`.
+This function handles both symbolic expressions and scalar values, tracking assumptions
+made during the computation.
+
+For scalar expressions, returns `(expr, sign(expr))`.
+For symbolic expressions, applies the Gruntz algorithm to compute the limit.
+
+# Arguments
+
+  - `expr`: The expression to take the limit of
+  - `x`: The variable approaching infinity
+  - `assumptions`: A set to track assumptions made during computation
+
+# Returns
+
+A tuple `(limit, sign)` where:
+
+  - `limit`: The computed limit value
+  - `sign`: The sign of the limit
+"""
+function signed_limit_inf(expr::Field, x::BasicSymbolic{Field}, assumptions) where {Field}
+    expr, sign(expr)
+end
+function signed_limit_inf(expr::BasicSymbolic{Field}, x::BasicSymbolic{Field}, assumptions) where {Field}
     expr === x && return (Inf, 1)
 
     Ω = most_rapidly_varying_subexpressions(expr, x, assumptions)
@@ -82,10 +115,10 @@ function signed_limit_inf(expr::BasicSymbolic{Field}, x::BasicSymbolic{Field}, a
         expr = log_exp_simplify(expr)
         # Ω = most_rapidly_varying_subexpressions(expr, x) NO! this line could lead to infinite recursion
         Ω = [log_exp_simplify(recursive(expr) do f, ex
-                ex isa BasicSymbolic{Field} || return ex
-                exprtype(ex) == SYM && return ex === x ? exp(x) : ex
-                operation(ex)(f.(arguments(ex))...)
-            end) for expr in Ω]
+                 ex isa BasicSymbolic{Field} || return ex
+                 exprtype(ex) == SYM && return ex === x ? exp(x) : ex
+                 operation(ex)(f.(arguments(ex))...)
+             end) for expr in Ω]
         ω_val = last(Ω)
     end
 
@@ -128,37 +161,98 @@ function signed_limit_inf(expr::BasicSymbolic{Field}, x::BasicSymbolic{Field}, a
     res
 end
 
+"""
+    recursive(f, args...)
+
+Apply function `f` recursively to its arguments, where `f` takes the recursive function as its first argument.
+
+This utility function enables recursive operations on nested expressions by passing the recursive
+function itself as the first argument to `f`.
+
+# Example
+
+```julia
+recursive(expr) do f, ex
+    # f is the recursive function, ex is the current expression
+    # Apply f to subexpressions and build the result
+end
+```
+"""
 function recursive(f, args...)
     g(args...) = f(g, args...)
     g(args...)
 end
 
-# TODO: use recursive or @rrule
+"""
+    log_exp_simplify(expr)
+
+Simplify expressions by canceling `log(exp(x))` → `x` transformations.
+
+This function performs basic log-exp simplifications that are always valid, specifically
+canceling `log(exp(x))` to `x`. This is a conservative simplification that doesn't extend domains.
+
+# Arguments
+
+  - `expr`: The expression to simplify
+
+# Returns
+
+The simplified expression with log-exp cancellations applied.
+"""
 log_exp_simplify(expr) = expr
 function log_exp_simplify(expr::BasicSymbolic)
     exprtype(expr) == SYM && return expr
-    exprtype(expr) == TERM && operation(expr) == log || return operation(expr)(log_exp_simplify.(arguments(expr))...)
+    exprtype(expr) == TERM && operation(expr) == log ||
+        return operation(expr)(log_exp_simplify.(arguments(expr))...)
     arg = log_exp_simplify(only(arguments(expr)))
     # TODO: return _log(arg)
-    arg isa BasicSymbolic && exprtype(arg) == TERM && operation(arg) == exp || return log(arg)
+    arg isa BasicSymbolic && exprtype(arg) == TERM && operation(arg) == exp ||
+        return log(arg)
     only(arguments(arg))
 end
 
-
-"""cancels log(exp(x)) and exp(log(x)), the latter may extend the domain"""
+"""
+cancels log(exp(x)) and exp(log(x)), the latter may extend the domain
+"""
 strong_log_exp_simplify(expr) = expr
 function strong_log_exp_simplify(expr::BasicSymbolic)
     exprtype(expr) == SYM && return expr
-    exprtype(expr) == TERM && operation(expr) in (log, exp) || return operation(expr)(strong_log_exp_simplify.(arguments(expr))...)
+    exprtype(expr) == TERM && operation(expr) in (log, exp) ||
+        return operation(expr)(strong_log_exp_simplify.(arguments(expr))...)
     arg = strong_log_exp_simplify(only(arguments(expr)))
     # TODO: return _log(arg)
-    arg isa BasicSymbolic && exprtype(arg) == TERM && operation(arg) in (log, exp) && operation(arg) != operation(expr) || return operation(expr)(arg)
+    arg isa BasicSymbolic && exprtype(arg) == TERM && operation(arg) in (log, exp) &&
+    operation(arg) != operation(expr) || return operation(expr)(arg)
     only(arguments(arg))
 end
 
-most_rapidly_varying_subexpressions(expr::Field, x::BasicSymbolic{Field}, assumptions) where Field = []
-function most_rapidly_varying_subexpressions(expr::BasicSymbolic{Field}, x::BasicSymbolic{Field}, assumptions) where Field
-    exprtype(x) == SYM || throw(ArgumentError("Must expand with respect to a symbol. Got $x"))
+"""
+    most_rapidly_varying_subexpressions(expr, x, assumptions)
+
+Find the most rapidly varying subexpressions (MRV set) in `expr` with respect to `x`.
+
+The MRV set consists of subexpressions that grow most rapidly as `x` approaches infinity.
+This is a key component of the Gruntz algorithm for computing symbolic limits.
+
+For scalar expressions, returns an empty list.
+
+# Arguments
+
+  - `expr`: The expression to analyze
+  - `x`: The variable with respect to which we find rapidly varying subexpressions
+  - `assumptions`: A set to track assumptions made during computation
+
+# Returns
+
+A list of the most rapidly varying subexpressions in `expr`.
+"""
+function most_rapidly_varying_subexpressions(expr::Field, x::BasicSymbolic{Field}, assumptions) where {Field}
+    []
+end
+function most_rapidly_varying_subexpressions(
+        expr::BasicSymbolic{Field}, x::BasicSymbolic{Field}, assumptions) where {Field}
+    exprtype(x) == SYM ||
+        throw(ArgumentError("Must expand with respect to a symbol. Got $x"))
     # TODO: this is slow. This whole algorithm is slow. Profile, benchmark, and optimize it.
     et = exprtype(expr)
     ret = if et == SYM
@@ -184,7 +278,8 @@ function most_rapidly_varying_subexpressions(expr::BasicSymbolic{Field}, x::Basi
             error("Not implemented: $op")
         end
     elseif et ∈ (ADD, MUL, DIV)
-        mapreduce(expr -> most_rapidly_varying_subexpressions(expr, x, assumptions), mrv_join(x, assumptions), arguments(expr), init=[])
+        mapreduce(expr -> most_rapidly_varying_subexpressions(expr, x, assumptions),
+            mrv_join(x, assumptions), arguments(expr), init = [])
     elseif et == POW
         args = arguments(expr)
         @assert length(args) == 2
@@ -195,22 +290,47 @@ function most_rapidly_varying_subexpressions(expr::BasicSymbolic{Field}, x::Basi
             error("Not implemented: POW with noninteger exponent $exponent. Transform to log/exp.")
         end
     else
-        error("Unknwon Expr type: $et")
+        error("Unknown Expr type: $et")
     end
     ret
 end
 
-is_exp_or_x(expr::BasicSymbolic, x::BasicSymbolic) =
+function is_exp_or_x(expr::BasicSymbolic, x::BasicSymbolic)
     expr === x || exprtype(expr) == TERM && operation(expr) == exp
+end
 
 """
-    f ≺ g iff log(f)/log(g) -> 0
+    compare_variance_rapidity(expr1, expr2, x, assumptions)
+
+Compare the rapidity of growth between two expressions as `x` approaches infinity.
+
+Returns:
+
+  - `-1` if `expr1` grows slower than `expr2` (i.e., `expr1 ≺ expr2`)
+  - `0` if `expr1` and `expr2` grow at the same rate (i.e., `expr1 ≍ expr2`)
+  - `1` if `expr1` grows faster than `expr2` (i.e., `expr1 ≻ expr2`)
+
+The comparison is based on the limit of `log(expr1)/log(expr2)` as `x` approaches infinity:
+
+  - `f ≺ g` iff `log(f)/log(g) → 0`
+  - `f ≍ g` iff `log(f)/log(g) → c` for some finite nonzero constant `c`
+  - `f ≻ g` iff `log(f)/log(g) → ∞`
+
+# Arguments
+
+  - `expr1`, `expr2`: Expressions to compare (must be of the form `x` or `exp(...)`).
+  - `x`: The variable approaching infinity
+  - `assumptions`: A set to track assumptions made during computation
+
+# Returns
+
+An integer (-1, 0, or 1) indicating the relative growth rates.
 """
-function compare_varience_rapidity(expr1, expr2, x, assumptions)
+function compare_variance_rapidity(expr1, expr2, x, assumptions)
     @assert is_exp_or_x(expr1, x)
     @assert is_exp_or_x(expr2, x)
     # expr1 === expr2 && return 0 # both x (or both same exp, either way okay, but for sure we cover the both x case)
-    # expr1 === x && expr2 !== x && return compare_varience_rapidity(expr2, expr1, x)
+    # expr1 === x && expr2 !== x && return compare_variance_rapidity(expr2, expr1, x)
     # @assert expr1 !== x
     # if expr2 !== x
     #     # they are both exp's, so it's safe (i.e. not a larger sub-expression) to call
@@ -238,22 +358,43 @@ function mrv_join(x, assumptions)
     function (mrvs1, mrvs2)
         isempty(mrvs1) && return mrvs2
         isempty(mrvs2) && return mrvs1
-        cmp = compare_varience_rapidity(first(mrvs1), first(mrvs2), x, assumptions)
+        cmp = compare_variance_rapidity(first(mrvs1), first(mrvs2), x, assumptions)
         if cmp == -1
             mrvs2
         elseif cmp == 1
             mrvs1
         else
-            vcat(mrvs1, mrvs2) # sets? unions? performance? nah. This saves us a topl-sort.
+            vcat(mrvs1, mrvs2) # sets? unions? performance? nah. This saves us a topo-sort.
         end
     end
 end
 
 """
-rewrite `expr` in the form `Aω^c` such that `A` is less rapidly varying than `ω` and `c` is
-a real number. `ω` is a symbol that represents `exp(h)`.
+    rewrite(expr, ω, h, x, assumptions)
+
+Rewrite `expr` in the form `A⋅ω^c` where `A` is less rapidly varying than `ω` and `c` is a real number.
+
+This function takes an exponential expression and rewrites it in terms of a distinguished
+exponential subexpression `ω`. The symbol `ω` represents `exp(h)` where `h` is the argument
+of the most rapidly varying exponential.
+
+The rewriting follows the formula: if `expr = exp(s)` and `ω = exp(h)`, then we compute
+`c = lim(s/h, x, ∞)` and rewrite as `exp(s-c⋅h) ⋅ ω^c`.
+
+# Arguments
+
+  - `expr`: An exponential expression to rewrite (must be of the form `exp(...)`)
+  - `ω`: A symbol representing the distinguished exponential `exp(h)`
+  - `h`: The argument of the distinguished exponential
+  - `x`: The variable approaching infinity
+  - `assumptions`: A set to track assumptions made during computation
+
+# Returns
+
+An expression of the form `A⋅ω^c` where `A` is less rapidly varying than `ω`.
 """
-function rewrite(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h::BasicSymbolic{Field}, x::BasicSymbolic{Field}, assumptions) where Field
+function rewrite(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field},
+        h::BasicSymbolic{Field}, x::BasicSymbolic{Field}, assumptions) where {Field}
     @assert exprtype(expr) == TERM && operation(expr) == exp
     @assert exprtype(ω) == SYM
     @assert exprtype(x) == SYM
@@ -266,10 +407,32 @@ function rewrite(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h::BasicS
 end
 
 """
-ω is a symbol that represents the expression exp(h).
+    get_series_term(expr, ω, h, i, assumptions)
+
+Get the coefficient of `ω^i` in the series expansion of `expr` around the most rapidly varying term.
+
+This function computes the `i`-th coefficient in the series expansion of `expr` in powers of `ω`,
+where `ω` represents the expression `exp(h)`. The expansion takes the form:
+`expr = c₀ + c₁⋅ω + c₂⋅ω² + ...`
+
+This is a core component of the Gruntz algorithm for computing limits.
+
+# Arguments
+
+  - `expr`: The expression to expand
+  - `ω`: A symbol representing the distinguished exponential `exp(h)`
+  - `h`: The argument of the distinguished exponential
+  - `i`: The power of `ω` for which to compute the coefficient
+  - `assumptions`: A set to track assumptions made during computation
+
+# Returns
+
+The coefficient of `ω^i` in the series expansion of `expr`.
 """
-function get_series_term(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h, i::Int, assumptions) where Field
-    exprtype(ω) == SYM || throw(ArgumentError("Must expand with respect to a symbol. Got $ω"))
+function get_series_term(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field},
+        h, i::Int, assumptions) where {Field}
+    exprtype(ω) == SYM ||
+        throw(ArgumentError("Must expand with respect to a symbol. Got $ω"))
 
     et = exprtype(expr)
     if et == SYM
@@ -292,7 +455,8 @@ function get_series_term(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h
                 for k in 1:i # the sum starts at 1
                     term = i ÷ k
                     if term * k == i # integral
-                        sm += (-get_series_term(arg, ω, h, term+exponent, assumptions)/t0)^k/k
+                        sm += (-get_series_term(
+                            arg, ω, h, term+exponent, assumptions)/t0)^k/k
                     end
                 end
                 # TODO: All these for loops are ugly and error-prone
@@ -310,16 +474,17 @@ function get_series_term(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h
                 for k in 1:i
                     term = i ÷ k
                     if term * k == i # integral
-                        sm += get_series_term(arg, ω, h, term, assumptions)^k/factorial(k) # this could overflow... oh well. It'l error if it does.
+                        sm += get_series_term(arg, ω, h, term, assumptions)^k/factorial(k) # this could overflow... oh well. It'll error if it does.
                     end
                 end
                 sm * exp(get_series_term(arg, ω, h, exponent, assumptions))
-            else @assert exponent > 0 # from the theory.
+            else
+                @assert exponent > 0 # from the theory.
                 # sum s^k/k!
                 for k in 1:i
                     term = i ÷ k
                     if term * k == i && term >= exponent # integral and not structural zero
-                        sm += get_series_term(arg, ω, h, term, assumptions)^k/factorial(k) # this could overflow... oh well. It'l error if it does.
+                        sm += get_series_term(arg, ω, h, term, assumptions)^k/factorial(k) # this could overflow... oh well. It'll error if it does.
                     end
                 end
                 sm
@@ -335,7 +500,7 @@ function get_series_term(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h
         exponent1 = get_leading_exponent(arg1, ω, h, assumptions)
         exponent2 = get_leading_exponent(arg2, ω, h, assumptions)
         sm = zero(Field)
-        for j in exponent1:(i-exponent2)
+        for j in exponent1:(i - exponent2)
             t1 = get_series_term(arg1, ω, h, j, assumptions)
             t2 = get_series_term(arg2, ω, h, i-j, assumptions)
             sm += t1 * t2
@@ -364,7 +529,7 @@ function get_series_term(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h
         den_leading_term = get_series_term(den, ω, h, den_exponent, assumptions)
         @assert !zero_equivalence(den_leading_term, assumptions)
         sm = zero(Field)
-        for j in num_exponent:i+den_exponent
+        for j in num_exponent:(i + den_exponent)
             t_num = get_series_term(num, ω, h, j, assumptions)
             exponent = i+den_exponent-j
             # TODO: refactor this to share code for the "sum of powers of a series" form
@@ -372,23 +537,49 @@ function get_series_term(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h
             for k in 1:exponent
                 term = exponent ÷ k
                 if term * k == exponent # integral
-                    sm2 += (-get_series_term(den, ω, h, term+den_exponent, assumptions)/den_leading_term)^k
+                    sm2 += (-get_series_term(
+                        den, ω, h, term+den_exponent, assumptions)/den_leading_term)^k
                 end
             end
             sm += sm2 * t_num
         end
         sm / den_leading_term
     else
-        error("Unknwon Expr type: $et")
+        error("Unknown Expr type: $et")
     end
 end
-function get_series_term(expr::Field, ω::BasicSymbolic{Field}, h, i::Int, assumptions) where Field
-    exprtype(ω) == SYM || throw(ArgumentError("Must expand with respect to a symbol. Got $ω"))
+function get_series_term(
+        expr::Field, ω::BasicSymbolic{Field}, h, i::Int, assumptions) where {Field}
+    exprtype(ω) == SYM ||
+        throw(ArgumentError("Must expand with respect to a symbol. Got $ω"))
     i == 0 ? expr : zero(Field)
 end
 
-function get_leading_exponent(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h, assumptions) where Field
-    exprtype(ω) == SYM || throw(ArgumentError("Must expand with respect to a symbol. Got $ω"))
+"""
+    get_leading_exponent(expr, ω, h, assumptions)
+
+Find the leading (smallest) exponent in the series expansion of `expr` in powers of `ω`.
+
+This function determines the power of the leading term in the series expansion:
+`expr = c_{e}⋅ω^e + c_{e+1}⋅ω^{e+1} + ...`
+where `e` is the leading exponent (the smallest exponent with a non-zero coefficient).
+
+Returns `Inf` if `expr` is equivalent to zero.
+
+# Arguments
+
+  - `expr`: The expression to analyze
+  - `ω`: A symbol representing the distinguished exponential `exp(h)`
+  - `h`: The argument of the distinguished exponential
+  - `assumptions`: A set to track assumptions made during computation
+
+# Returns
+
+The leading exponent `e`, or `Inf` if `expr` is equivalent to zero.
+"""
+function get_leading_exponent(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Field}, h, assumptions) where {Field}
+    exprtype(ω) == SYM ||
+        throw(ArgumentError("Must expand with respect to a symbol. Got $ω"))
 
     zero_equivalence(expr, assumptions) && return Inf
 
@@ -410,7 +601,9 @@ function get_leading_exponent(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Fiel
             else
                 # There will never be a term with power less than 0, and the zero power term
                 # is log(T0) which is handled above with the "isone" check.
-                findfirst(i -> zero_equivalence(get_series_term(expr, ω, h, i, assumptions), assumptions), 1:typemax(Int))
+                findfirst(
+                    i -> zero_equivalence(get_series_term(expr, ω, h, i, assumptions), assumptions),
+                    1:typemax(Int))
             end
         elseif op == exp
             0
@@ -418,7 +611,8 @@ function get_leading_exponent(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Fiel
             error("Not implemented: $op")
         end
     elseif et == ADD
-        exponent = minimum(get_leading_exponent(arg, ω, h, assumptions) for arg in  arguments(expr))
+        exponent = minimum(get_leading_exponent(arg, ω, h, assumptions)
+        for arg in arguments(expr))
         for i in exponent:typemax(Int)
             sm = sum(get_series_term(arg, ω, h, i, assumptions) for arg in arguments(expr))
             if !zero_equivalence(sm, assumptions)
@@ -441,14 +635,16 @@ function get_leading_exponent(expr::BasicSymbolic{Field}, ω::BasicSymbolic{Fiel
         args = arguments(expr)
         @assert length(args) == 2
         num, den = args
-         # The naive answer is actually correct. See the get_series_term implementation for how.
-        get_leading_exponent(num, ω, h, assumptions) - get_leading_exponent(den, ω, h, assumptions)
+        # The naive answer is actually correct. See the get_series_term implementation for how.
+        get_leading_exponent(num, ω, h, assumptions) -
+        get_leading_exponent(den, ω, h, assumptions)
     else
-        error("Unknwon Expr type: $et")
+        error("Unknown Expr type: $et")
     end
 end
-function get_leading_exponent(expr::Field, ω::BasicSymbolic{Field}, h, assumptions) where Field
-    exprtype(ω) == SYM || throw(ArgumentError("Must expand with respect to a symbol. Got $ω"))
+function get_leading_exponent(expr::Field, ω::BasicSymbolic{Field}, h, assumptions) where {Field}
+    exprtype(ω) == SYM ||
+        throw(ArgumentError("Must expand with respect to a symbol. Got $ω"))
     zero_equivalence(expr, assumptions) ? Inf : 0
 end
 
@@ -460,9 +656,33 @@ function _log(x::BasicSymbolic, ω, h)
     log(x)
 end
 
-"""Is `expr` zero on its domain?"""
+"""
+    zero_equivalence(expr, assumptions)
+
+Determine if `expr` is equivalent to zero on its domain using heuristic methods.
+
+This function uses symbolic simplification to determine if an expression is zero.
+Since zero equivalence of general expressions is undecidable, this function uses heuristics
+and records the result as an assumption that must hold for the computed limit to be correct.
+
+The function applies strong log-exp simplification and algebraic expansion before testing
+for zero equivalence.
+
+# Arguments
+
+  - `expr`: The expression to test for zero equivalence
+  - `assumptions`: A set to record assumptions made (updated in-place)
+
+# Returns
+
+`true` if the expression is determined to be zero, `false` otherwise.
+
+# Side Effects
+
+Adds an assumption about the zero-equivalence of `expr` to the `assumptions` set.
+"""
 function zero_equivalence(expr, assumptions)
-    res = iszero(simplify(strong_log_exp_simplify(expr), expand=true)) === true
+    res = iszero(simplify(strong_log_exp_simplify(expr), expand = true)) === true
     push!(assumptions, res ? iszero(expr) : !iszero(expr))
     res
 end
